s;; TODO: make a version of fill-buffer optionally be able to
;;       do a non-blocking read to 'try' to fill the buffer
;; TODO: since this is a buffered stream, be able to putback data as
;; well as read it.
;; TODO: don't rewrite the entries from cl-fad, just use it

(in-package :krb-util)
; (cl:in-package :krb-util)

(defvar *default-line-terminator* "\\r?\\n"
  "Default line terminator, a regex: an optional CR followed by an LF")

(defvar *default-buffer-size* (* 8 1024)
  "The default read/write buffer size.")

(defvar *buffer-size* *default-buffer-size*
  "Default size of the buffer when attempting to perform bulk IO.")

(defclass buffered-stream ()
  ((read-buffer
    :accessor read-buffer
    :initform (make-array 0)
    :initarg :buffer
    :documentation "The bulk read buffer.")
   (buffer-size
    :accessor buffer-size
    :initform 0
    :initarg :buffer-size
    :documentation "The size of the read-buffer, this may change as line lengths force it to grow.")
   (buffer-end-position
    :accessor buffer-end-position
    :initform 0
    :initarg :buffer-end-position
    :documentation "Current position in the buffer for next read.")
   (line-terminator
    :accessor line-terminator
    :initform *default-line-terminator*
    :initarg :eol
    :documentation "")
   (at-end-of-file
    :accessor at-end-of-file
    :initform nil
    :initarg :at-eol
    :documentation "The stream has hit EOF, no more can be read.")
   (input-stream
    :accessor input-stream
    :initform nil
    :initarg :stream
    :documentation "The input stream being operated on.")))

(defgeneric set-buffer-size (reader size))

(defmethod set-buffer-size ((reader buffered-stream) size)
  "Sets the buffer size in the reader - the buffer size represents the
amount of data in the buffer, not the total capacity."
  (setf (buffer-size reader) size))

(defgeneric peek-read-buffer (reader))

(defmethod peek-read-buffer ((reader buffered-stream))
  (subseq (read-buffer reader)
	  0
	  (buffer-end-position reader)))

(defun make-buffered-stream (stream &key read-buffer buffer-size line-terminator)
  "Create a buffered stream, wrapping the given stream."
  (unless buffer-size
    (setf buffer-size (get-read-buffer-size)))
  (unless read-buffer
    ;; skip using the fill-pointer for now, just realloc the buffer to
    ;; make this simple...come back and chase efficency later
    ;; (setf read-buffer (make-array buffer-size
    ;; 				  :fill-pointer 0
    ;; 				  :element-type 'character))
    (setf read-buffer (make-array buffer-size
				  :element-type 'character)))
  (unless line-terminator
    (setf line-terminator *default-line-terminator*))
  (let ((rdr (make-instance 'buffered-stream)))
    (setf (read-buffer rdr) read-buffer)
    (setf (buffer-size rdr) buffer-size)
    (setf (line-terminator rdr) line-terminator)
    (setf (input-stream rdr) stream)
    rdr))

(defun set-read-buffer-size (size)
  "Sets the default buffer capacity (global) to the given value."
  (setf *buffer-size* size))

(defun get-read-buffer-size ()
  "Retrieve the read buffer size (capacity, this is a global)."
  *buffer-size*)

;; use bulk-io (read-sequence using a vector), find the next instance
;; of line-terminator (as a regular expression), use file-position on
;; the stream to reset it back to just after the line-terminator,
;; return the full line, with the terminator still on the end of the
;; string...this is all basically because I want to be able to
;; controll what the line terminator is and I don't know how to do
;; that with read-line.

(defgeneric fill-buffer (reader))

(defmethod fill-buffer ((reader buffered-stream))
  (when (= (buffer-end-position reader)
	 (buffer-size reader))
    ;;(format t "fill-buffer: buffer at limit, unable to fill more~&")
    (return-from fill-buffer nil))
  (let ((end-of-read (read-sequence  (read-buffer reader) (input-stream reader)
				   :start (buffer-end-position reader)
				   :end (buffer-size reader))))
;;     (format t "fill-buffer: tried additional read (buffer-size:~a, buffer-end-position:~a), end-of-read:~a~&"
;; 	    (buffer-size reader)
;; 	    (buffer-end-position reader)
;; 	    end-of-read)
    (if (< end-of-read (buffer-size reader))
	(setf (at-end-of-file reader) t))
    (cond ((= end-of-read (buffer-end-position reader)) ;; nothing was read
	   nil)
	  (t
	   (setf (buffer-end-position reader) end-of-read)
	   t))))


(defgeneric find-line-terminator (reader))

(defmethod find-line-terminator ((reader buffered-stream))
;;   (format t "find-line-terminator: looking for '~a' in '~a' before ~a~&"
;; 	  (line-terminator reader)
;; 	  (read-buffer reader)
;; 	  (buffer-end-position reader))
  (cl-ppcre:scan (line-terminator reader)
		 (read-buffer reader)
		 :end (buffer-end-position reader)))

(defgeneric consume-next-line (reader))

(defmethod consume-next-line ((reader buffered-stream))
  (destructuring-bind (match-start match-end reg-starts reg-ends)
      (multiple-value-list (find-line-terminator reader))
    (declare (ignore reg-starts match-start reg-ends))
    (aprog1
	(subseq (read-buffer reader) 0 match-end)
      ;; (format t "consume-next-line: match-end:~a line:'~a', buffer-end-pos:~a~&" match-end it (buffer-end-position reader))
      (replace (read-buffer reader)
	       (subseq (read-buffer reader)
		       match-end))
      (decf (buffer-end-position reader) match-end))))

(defgeneric consume-remaining-buffer (reader))

(defmethod consume-remaining-buffer ((reader buffered-stream))
  (aprog1
      (subseq (read-buffer reader) 0 (buffer-end-position reader))
    (setf (buffer-end-position reader) 0)))

(defgeneric buffer-empty-p (reader))

(defmethod buffer-empty-p ((reader buffered-stream))
  (<= (buffer-end-position reader) 0))

(defgeneric read-next-line (reader))

(defmethod read-next-line ((reader buffered-stream))
  (cond
    ((and (not (buffer-empty-p reader))
	  (at-end-of-file reader)
	  (find-line-terminator reader))
     (format t "read-next-line: at eof, buff not empty, has line-term, returning next line~&")
     (consume-next-line reader))
    ((and (not (buffer-empty-p reader))
	  (at-end-of-file reader))
     (format t "read-next-line: at eof, buff not empty, returning remainder~&")
     (consume-remaining-buffer reader))
    ((at-end-of-file reader)
     (format t "read-next-line: at eof, buffer is empty all done (nil)~&")
     nil)
    ;; not at EOF, see if there is a 'line' in the buffer
    ((find-line-terminator reader)
     (format t "read-next-line: found terminator, consuming line...~&")
     (consume-next-line reader))
    ;; buffer was empty or had incomplete line, try filling and finding a line
    ((and (fill-buffer reader)
	  (find-line-terminator reader))
     (format t "read-next-line: found terminator after fill-buffer, consuming line...~&")
     (consume-next-line reader))
    ;; we performed a read, but the buffer wasn't big enough (no line
    ;; found), increase it and try again
    (t
     (let* ((new-size (* 2 (buffer-size reader)))
	   (new-buffer (make-array new-size :element-type 'character)))
     (format t "read-next-line: buffer(~a,end:~a) too small, reallocating from ~a to ~a~&"
	     (subseq (read-buffer reader)
		     0 (min 100 (buffer-size reader)))
	     (buffer-end-position reader)
	     (buffer-size reader)
	     new-size)
     (replace new-buffer (read-buffer reader))
     (format t "read-next-line: after replace, new-buffer:~a~&"
	     (subseq new-buffer 0 (min 100 new-size)))
     (setf (read-buffer reader) new-buffer)
     (setf (buffer-size reader) new-size)
    (format t "read-next-line: recursing to fill the buffer again...~&")
     (read-next-line reader)))))

  
(defgeneric drain-stream (reader))

(defmethod drain-stream ((reader buffered-stream))
  (cond
    ;; finally at EOF, return whatever's in the buffer
    ((at-end-of-file reader)
     (consume-remaining-buffer reader))
    ;; buffer was empty or had incomplete line, try filling, if
    ;; something is read, recurse
    ((fill-buffer reader)
     (drain-stream reader))
    ;; nothing was read, if not at EOF, increase buffer and recurse
    ((not (at-end-of-file reader))
     (let* ((new-size (* 2 (buffer-size reader)))
	    (new-buffer (make-array new-size :element-type 'character)))
       (replace new-buffer (read-buffer reader))
       (setf (read-buffer reader) new-buffer)
       (setf (buffer-size reader) new-size)
       (drain-stream reader)))
    ;; nothing was read, and we were at EOF, return whatever we had
    (t
     (consume-remaining-buffer reader))))

(defun stream->string (stream)
  (drain-stream (make-buffered-stream stream)))

(defun stream->lines (stream &optional (line-terminator *default-line-terminator*))
  (loop
     with rdr = (make-buffered-stream stream :line-terminator line-terminator)
     for line = (read-next-line rdr)
     while line
     collect line))


;; (defun write-to-stream ()
;;   (error "implement this"))

(defun file->string (inpath)
  (let* ((size (sb-posix:stat-size (sb-posix:stat inpath)))
         (buff (make-array size :element-type 'character)))
    (with-open-file (in inpath :direction :input)
      (read-sequence buff in))
    buff))

;;(length (file->string (krb-util:expand-tilde-file "~/iso.html")))

(defun file->lines (inpath)
  (with-open-file (in inpath :direction :input)
    (stream->lines in)))



;; defun url->string
;; defun append-file
;; defun write-file

(defun write-file (file &rest things)
  (with-open-file (out file :direction :output :if-exists :supersede)
    (loop for thing in things
       do
	 (format out "~a" thing))))

(defun list->file (file things)
  (with-open-file (out file :direction :output :if-exists :supersede)
    (loop for thing in things
       do
	 (format out "~a" thing))))

